use crate::{ast2::tag::AstNode, variant};

use super::{tag::AstNodeExt, *};

pub trait AstExt {
    fn get_node_children(&self, node: Index) -> Vec<Index>;
    fn get_node_tag_and_data(&self, node: Index) -> (Tag, Data);

    fn get_node_data_for_tag(&self, idx: Index, tag: Tag) -> Option<Data> {
        let (t, data) = self.get_node_tag_and_data(idx);
        if t == tag {
            Some(data)
        } else {
            None
        }
    }

    fn expect_node_data_for_tag(&self, idx: Index, tag: Tag) -> Data {
        self.get_node_data_for_tag(idx, tag)
            .expect(&format!("node {idx} is not a {tag:?}"))
    }
}

impl AstExt for &Ast {
    fn get_node_children(&self, node: Index) -> Vec<Index> {
        Ast::get_node_children(self, node)
    }

    fn get_node_tag_and_data(&self, node: Index) -> (Tag, Data) {
        (self.tags[node], self.datas[node])
    }
}

impl AstExt for &mut Ast {
    fn get_node_children(&self, node: Index) -> Vec<Index> {
        Ast::get_node_children(self, node)
    }

    fn get_node_tag_and_data(&self, node: Index) -> (Tag, Data) {
        (self.tags[node], self.datas[node])
    }
}

pub struct AstVisitor<AstT> {
    pub(super) ast: AstT,
    pub(super) scopes: Vec<Index>,
    pub(super) nodes: Vec<A>,
    pub(super) rev: bool,
}

impl<AstT> AstVisitor<AstT>
where
    AstT: AstExt,
{
    pub fn visit<
        F: FnMut(&mut AstT, &[Index], Index, Tag, Data),
        G: FnMut(&mut AstT, &[Index], Index, Tag, Data),
    >(
        &mut self,
        mut pre: F,
        mut post: G,
    ) {
        while let Some(node) = self.nodes.pop() {
            match node {
                A::PushChildren(i) => {
                    self.nodes.push(A::PopSelf(i));
                    let children_iter = self
                        .ast
                        .get_node_children(i)
                        .into_iter()
                        .map(|i| A::PushChildren(i));
                    // inverse because we are popping from the end
                    if !self.rev {
                        self.nodes.extend(children_iter.rev())
                    } else {
                        self.nodes.extend(children_iter)
                    };

                    let (tag, data) = self.ast.get_node_tag_and_data(i);

                    let _ = pre(&mut self.ast, &self.scopes, i, tag, data);

                    match tag {
                        Tag::File
                        | Tag::FunctionDecl
                        | Tag::GlobalDecl
                        | Tag::Block
                        | Tag::BlockTrailingExpr => {
                            self.scopes.push(i);
                        }
                        _ => {}
                    }
                }
                A::PopSelf(i) => {
                    // already popped.
                    let (tag, data) = self.ast.get_node_tag_and_data(i);

                    let _ = post(&mut self.ast, &self.scopes, i, tag, data);

                    match tag {
                        Tag::File
                        | Tag::FunctionDecl
                        | Tag::GlobalDecl
                        | Tag::Block
                        | Tag::BlockTrailingExpr => {
                            self.scopes.pop();
                        }
                        _ => {}
                    }
                }
            }
        }
    }
    pub fn visit_pre<F: FnMut(&mut AstT, &[Index], Index, Tag, Data)>(&mut self, mut cb: F) {
        while let Some(node) = self.nodes.pop() {
            match node {
                A::PushChildren(i) => {
                    self.nodes.push(A::PopSelf(i));
                    let children_iter = self
                        .ast
                        .get_node_children(i)
                        .into_iter()
                        .map(|i| A::PushChildren(i));
                    // inverse because we are popping from the end
                    if !self.rev {
                        self.nodes.extend(children_iter.rev())
                    } else {
                        self.nodes.extend(children_iter)
                    };

                    let (tag, data) = self.ast.get_node_tag_and_data(i);

                    let _ = cb(&mut self.ast, &self.scopes, i, tag, data);

                    match tag {
                        Tag::File
                        | Tag::FunctionDecl
                        | Tag::GlobalDecl
                        | Tag::Block
                        | Tag::BlockTrailingExpr => {
                            self.scopes.push(i);
                        }
                        _ => {}
                    }
                }
                A::PopSelf(i) => {
                    // already popped.
                    let (tag, _data) = self.ast.get_node_tag_and_data(i);

                    match tag {
                        Tag::File
                        | Tag::FunctionDecl
                        | Tag::GlobalDecl
                        | Tag::Block
                        | Tag::BlockTrailingExpr => {
                            self.scopes.pop();
                        }
                        _ => {}
                    }
                }
            }
        }
    }
    pub fn visit_post<F: FnMut(&mut AstT, &[Index], Index, Tag, Data)>(&mut self, mut cb: F) {
        while let Some(node) = self.nodes.pop() {
            match node {
                A::PushChildren(i) => {
                    self.nodes.push(A::PopSelf(i));
                    let children_iter = self
                        .ast
                        .get_node_children(i)
                        .into_iter()
                        .map(|i| A::PushChildren(i));
                    if self.rev {
                        self.nodes.extend(children_iter.rev())
                    } else {
                        self.nodes.extend(children_iter)
                    };

                    let (tag, _data) = self.ast.get_node_tag_and_data(i);

                    match tag {
                        Tag::File
                        | Tag::FunctionDecl
                        | Tag::GlobalDecl
                        | Tag::Block
                        | Tag::BlockTrailingExpr => {
                            self.scopes.push(i);
                        }
                        _ => {}
                    }
                }
                A::PopSelf(i) => {
                    // already popped.
                    let (tag, data) = self.ast.get_node_tag_and_data(i);

                    let _ = cb(&mut self.ast, &self.scopes, i, tag, data);

                    match tag {
                        Tag::File
                        | Tag::FunctionDecl
                        | Tag::GlobalDecl
                        | Tag::Block
                        | Tag::BlockTrailingExpr => {
                            self.scopes.pop();
                        }
                        _ => {}
                    }
                }
            }
        }
    }
}

pub enum A {
    PushChildren(Index),
    PopSelf(Index),
}

impl Ast {
    pub fn visitor_mut(&mut self) -> AstVisitor<&mut Self> {
        AstVisitor {
            scopes: vec![],
            nodes: self
                .get_root_file_indices()
                .map(|i| visitor::A::PushChildren(i))
                .collect(),
            ast: self,
            rev: false,
        }
    }
    pub fn visitor_rev_mut(&mut self) -> AstVisitor<&mut Self> {
        AstVisitor {
            scopes: vec![],
            nodes: self
                .get_root_file_indices()
                .map(|i| visitor::A::PushChildren(i))
                .collect(),
            ast: self,
            rev: true,
        }
    }

    pub fn visitor(&self) -> AstVisitor<&Self> {
        AstVisitor {
            scopes: vec![],
            nodes: self
                .get_root_file_indices()
                .map(|i| visitor::A::PushChildren(i))
                .collect(),
            ast: self,
            rev: false,
        }
    }
}

macro_rules! tag_visit_fn {
    ($($tag:tt {$($field_name:ident : $field_ty:ty),* $(,)?}),* $(,)?) => {
        $(
            paste::paste! {

                fn [<visit_ $tag:snake>](&mut self, ast: Ast, idx: Index) -> Result<Self::Value, Self::Error> {
                    variant!(ast.get_ast_node(idx) => AstNode::$tag { $($field_name),* });
                    self.[<visit_ $tag:snake _impl>](ast, idx, $($field_name),*)
                }
                fn [<visit_ $tag:snake _impl>](&mut self, ast: Ast, idx: Index, $($field_name: $field_ty),*) -> Result<Self::Value, Self::Error> {
                    _ = (ast, idx, $($field_name),*);
                    Err(Self::UNIMPL)
                }
            }
        )*
    };
}

use intern::Index as Interned;

pub trait AstVisitorTrait<Ast: AstExt + AstNodeExt> {
    type Error;
    type Value;
    const UNIMPL: Self::Error;

    tag_visit_fn!(
        File {
            decls: Vec<Index>,
        },
        FunctionProto {
            name: Interned,
            return_type: Index,
            parameter_list: Index,
        },
        FunctionProtoInterned {
            name: Interned,
            ty: Interned,
        },
        FunctionDecl {
            proto: Index,
            body: Index,
        },
        ParameterList {
            params: Vec<Index>,
        },
        Parameter {
            ty: Index,
            name: Interned,
        },
        Block {
            statements: Vec<Index>,
            expr: Option<Index>,
        },
        Constant {
            ty: Index,
            value: Interned,
        },
        ExprStmt {
            expr: Index,
        },
        ReturnStmt {},
        ReturnExprStmt {
            expr: Index,
        },
        VarDecl {
            name: Interned,
            ty: Index,
        },
        MutVarDecl {
            name: Interned,
            ty: Index,
        },
        VarDeclAssignment {
            name: Interned,
            expr: Index,
            ty: Option<Index>,
        },
        MutVarDeclAssignment {
            name: Interned,
            expr: Index,
            ty: Option<Index>,
        },
        GlobalDecl {
            name: Interned,
            expr: Index,
            ty: Index,
        },
        StructDecl {
            name: Interned,
            flags: StructFlags,
            field_names: Vec<Interned>,
            field_types: Vec<Index>,
        },
        StructDeclInterned {
            name: Interned,
            ty: Interned,
        },
        FieldDecl {
            name: Interned,
            ty: Index,
        },
        DeclRef {
            decl: Index,
        },
        DeclRefUnresolved {
            scope: Index,
            name: Interned,
        },
        InternedType {
            intern: Interned,
        },
        TypeDeclRef {
            decl: Index,
        },
        TypeDeclRefUnresolved {
            scope: Index,
            name: Interned,
        },
        PointerType {
            ty: Index,
            flags: PointerFlags,
        },
        ArrayType {
            length: Index,
            pointer: Index,
        },
        CallExpr {
            func: Index,
            argument_list: Index,
        },
        FieldAccess {
            field_name: Interned,
            expr: Index,
        },
        ArgumentList {
            arguments: Vec<Index>,
        },
        Argument {
            expr: Index,
            name: Option<Interned>,
        },
        ExplicitCast {
            expr: Index,
            ty: Index,
        },
        Deref {
            expr: Index,
        },
        AddressOf {
            expr: Index,
        },
        Not {
            expr: Index,
        },
        Negate {
            expr: Index,
        },
        PlaceToValueConversion {
            expr: Index,
        },
        ValueToPlaceConversion {
            expr: Index,
        },
        Or {
            lhs: Index,
            rhs: Index,
        },
        And {
            lhs: Index,
            rhs: Index,
        },
        BitOr {
            lhs: Index,
            rhs: Index,
        },
        BitXOr {
            lhs: Index,
            rhs: Index,
        },
        BitAnd {
            lhs: Index,
            rhs: Index,
        },
        Eq {
            lhs: Index,
            rhs: Index,
        },
        NEq {
            lhs: Index,
            rhs: Index,
        },
        Lt {
            lhs: Index,
            rhs: Index,
        },
        Gt {
            lhs: Index,
            rhs: Index,
        },
        Le {
            lhs: Index,
            rhs: Index,
        },
        Ge {
            lhs: Index,
            rhs: Index,
        },
        Shl {
            lhs: Index,
            rhs: Index,
        },
        Shr {
            lhs: Index,
            rhs: Index,
        },
        Add {
            lhs: Index,
            rhs: Index,
        },
        Sub {
            lhs: Index,
            rhs: Index,
        },
        Mul {
            lhs: Index,
            rhs: Index,
        },
        Div {
            lhs: Index,
            rhs: Index,
        },
        Rem {
            lhs: Index,
            rhs: Index,
        },
        Assign {
            lhs: Index,
            rhs: Index,
        },
        SubscriptExpr {
            lhs: Index,
            index: Index,
        },
        IfExpr {
            cond: Index,
            body: Index,
        },
        IfElseExpr {
            cond: Index,
            a: Index,
            b: Index,
        },
        Error {
            err: ParseError,
        },
        Undefined {},
    );

    fn visit_any(&mut self, ast: Ast, idx: Index) -> Result<Self::Value, Self::Error> {
        match ast.get_ast_node(idx) {
            AstNode::File { decls } => self.visit_file_impl(ast, idx, decls),
            AstNode::FunctionProto {
                name,
                return_type,
                parameter_list,
            } => self.visit_function_proto_impl(ast, idx, name, return_type, parameter_list),
            AstNode::FunctionProtoInterned { name, ty } => {
                self.visit_function_proto_interned_impl(ast, idx, name, ty)
            }
            AstNode::FunctionDecl { proto, body } => {
                self.visit_function_decl_impl(ast, idx, proto, body)
            }
            AstNode::ParameterList { params } => self.visit_parameter_list_impl(ast, idx, params),
            AstNode::Parameter { ty, name } => self.visit_parameter_impl(ast, idx, ty, name),
            AstNode::Block { statements, expr } => {
                self.visit_block_impl(ast, idx, statements, expr)
            }
            AstNode::Constant { ty, value } => self.visit_constant_impl(ast, idx, ty, value),
            AstNode::ExprStmt { expr } => self.visit_expr_stmt_impl(ast, idx, expr),
            AstNode::ReturnStmt => self.visit_return_stmt_impl(ast, idx),
            AstNode::ReturnExprStmt { expr } => self.visit_return_expr_stmt_impl(ast, idx, expr),
            AstNode::VarDecl { name, ty } => self.visit_var_decl_impl(ast, idx, name, ty),
            AstNode::MutVarDecl { name, ty } => self.visit_mut_var_decl_impl(ast, idx, name, ty),
            AstNode::VarDeclAssignment { name, expr, ty } => {
                self.visit_var_decl_assignment_impl(ast, idx, name, expr, ty)
            }
            AstNode::MutVarDeclAssignment { name, expr, ty } => {
                self.visit_mut_var_decl_assignment_impl(ast, idx, name, expr, ty)
            }
            AstNode::GlobalDecl { name, expr, ty } => {
                self.visit_global_decl_impl(ast, idx, name, expr, ty)
            }
            AstNode::StructDecl {
                name,
                flags,
                field_names,
                field_types,
            } => self.visit_struct_decl_impl(ast, idx, name, flags, field_names, field_types),
            AstNode::StructDeclInterned { name, ty } => {
                self.visit_struct_decl_interned_impl(ast, idx, name, ty)
            }
            AstNode::FieldDecl { name, ty } => self.visit_field_decl_impl(ast, idx, name, ty),
            AstNode::DeclRef { decl } => self.visit_decl_ref_impl(ast, idx, decl),
            AstNode::DeclRefUnresolved { scope, name } => {
                self.visit_decl_ref_unresolved_impl(ast, idx, scope, name)
            }
            AstNode::InternedType { intern } => self.visit_interned_type_impl(ast, idx, intern),
            AstNode::TypeDeclRef { decl } => self.visit_type_decl_ref_impl(ast, idx, decl),
            AstNode::TypeDeclRefUnresolved { scope, name } => {
                self.visit_type_decl_ref_unresolved_impl(ast, idx, scope, name)
            }
            AstNode::PointerType { ty, flags } => self.visit_pointer_type_impl(ast, idx, ty, flags),
            AstNode::ArrayType { length, pointer } => {
                self.visit_array_type_impl(ast, idx, length, pointer)
            }
            AstNode::CallExpr {
                func,
                argument_list,
            } => self.visit_call_expr_impl(ast, idx, func, argument_list),
            AstNode::FieldAccess { field_name, expr } => {
                self.visit_field_access_impl(ast, idx, field_name, expr)
            }
            AstNode::ArgumentList { arguments } => {
                self.visit_argument_list_impl(ast, idx, arguments)
            }
            AstNode::Argument { expr, name } => self.visit_argument_impl(ast, idx, expr, name),
            AstNode::ExplicitCast { expr, ty } => self.visit_explicit_cast_impl(ast, idx, expr, ty),
            AstNode::Deref { expr } => self.visit_deref_impl(ast, idx, expr),
            AstNode::AddressOf { expr } => self.visit_address_of_impl(ast, idx, expr),
            AstNode::Not { expr } => self.visit_not_impl(ast, idx, expr),
            AstNode::Negate { expr } => self.visit_negate_impl(ast, idx, expr),
            AstNode::PlaceToValueConversion { expr } => {
                self.visit_place_to_value_conversion_impl(ast, idx, expr)
            }
            AstNode::ValueToPlaceConversion { expr } => {
                self.visit_value_to_place_conversion_impl(ast, idx, expr)
            }
            AstNode::Or { lhs, rhs } => self.visit_or_impl(ast, idx, lhs, rhs),
            AstNode::And { lhs, rhs } => self.visit_and_impl(ast, idx, lhs, rhs),
            AstNode::BitOr { lhs, rhs } => self.visit_bit_or_impl(ast, idx, lhs, rhs),
            AstNode::BitXOr { lhs, rhs } => self.visit_bit_x_or_impl(ast, idx, lhs, rhs),
            AstNode::BitAnd { lhs, rhs } => self.visit_bit_and_impl(ast, idx, lhs, rhs),
            AstNode::Eq { lhs, rhs } => self.visit_eq_impl(ast, idx, lhs, rhs),
            AstNode::NEq { lhs, rhs } => self.visit_n_eq_impl(ast, idx, lhs, rhs),
            AstNode::Lt { lhs, rhs } => self.visit_lt_impl(ast, idx, lhs, rhs),
            AstNode::Gt { lhs, rhs } => self.visit_gt_impl(ast, idx, lhs, rhs),
            AstNode::Le { lhs, rhs } => self.visit_le_impl(ast, idx, lhs, rhs),
            AstNode::Ge { lhs, rhs } => self.visit_ge_impl(ast, idx, lhs, rhs),
            AstNode::Shl { lhs, rhs } => self.visit_shl_impl(ast, idx, lhs, rhs),
            AstNode::Shr { lhs, rhs } => self.visit_shr_impl(ast, idx, lhs, rhs),
            AstNode::Add { lhs, rhs } => self.visit_add_impl(ast, idx, lhs, rhs),
            AstNode::Sub { lhs, rhs } => self.visit_sub_impl(ast, idx, lhs, rhs),
            AstNode::Mul { lhs, rhs } => self.visit_mul_impl(ast, idx, lhs, rhs),
            AstNode::Div { lhs, rhs } => self.visit_div_impl(ast, idx, lhs, rhs),
            AstNode::Rem { lhs, rhs } => self.visit_rem_impl(ast, idx, lhs, rhs),
            AstNode::Assign { lhs, rhs } => self.visit_assign_impl(ast, idx, lhs, rhs),
            AstNode::SubscriptExpr { lhs, index: rhs } => {
                self.visit_subscript_expr_impl(ast, idx, lhs, rhs)
            }
            AstNode::IfExpr { cond, body } => self.visit_if_expr_impl(ast, idx, cond, body),
            AstNode::IfElseExpr { cond, a, b } => {
                self.visit_if_else_expr_impl(ast, idx, cond, a, b)
            }
            AstNode::Error { err } => self.visit_error_impl(ast, idx, err),
            AstNode::Undefined => self.visit_undefined_impl(ast, idx),
            AstNode::Root { .. } => unreachable!(),
        }
    }
}
